#!/usr/bin/env python3

"""
Command-line entry point
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111,w1202,r1705
# pylint: disable=logging-fstring-interpolation, fixme, import-outside-toplevel


# DON'T import logger globally here
# we need to be able to mess inside the logger module
# before it gets loaded from another way

import time
import os
from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging
import asyncio


from pkg_resources import resource_string #, resource_exists, resource_filename

from asynciojobs import Scheduler, Job

from asyncssh.logging import set_log_level as asyncssh_set_log_level

# apparently using relative imports is - at least that's what pylint says
# defining global variables like 'config'
# pylint: disable = redefined-outer-name
from .config import Config
from .imagesrepo import ImagesRepo
from .selector import Selector, add_selector_arguments, selected_selector
from .action import Action
from .display import Display
from .display_curses import DisplayCurses
from .node import Node
from .imageloader import ImageLoader
from .imagesaver import ImageSaver
from .monitor.loop import MonitorLoop
from .monitor.nodes import MonitorNodes
from .monitor.phones import MonitorPhones
from .monitor.pdus import MonitorPdus
from .monitor.leases import MonitorLeases
from .monitor.accountsmanager import AccountsManager
from .ssh import SshProxy
from .leases import Leases
from .book import Book
from .inventorynodes import InventoryNodes
from .inventoryphones import InventoryPhones
from .inventorypdus import InventoryPdus


# a supported command comes with a driver function
# in this module, that takes a list of args
# and returns 0 for success and s/t else otherwise
# specifically, command
# rhubarbe load -i fedora 12
# would result in a call
# load ( "-i", "fedora", "12" )


####################
RESERVATION_REQUIRED = "This function requires a valid reservation - "\
                       "or to be root or a privileged user"


####################
# exposed to the outside world (typically r2lab's nightly)

def check_reservation(leases, *,                        # pylint: disable=w0621
                      root_allowed=True, verbose=False, login=None):
    """
    return a bool indicating if a login currently has a lease

    if login is None, the current login is used instead

    when True, root_allowed means that the root user is always granted access

    verbose can be
    None  : does not write anything
    False : write a message if lease is not there
    True : always write a message
    """

    async def check_leases():
        # login = None means use my login
        # we don't use booked_now_by_me because of the verbose message
        # we use another variable to avoid using nonlocal
        actual_login = login or leases.login
        if verbose:
            print(f"Checking current reservation for {actual_login} : ", end="")
        is_fine = await leases.booked_now_by(login=actual_login,
                                             root_allowed=root_allowed)
        if is_fine:
            if verbose:
                print("OK")
        else:
            if verbose is None:
                pass
            elif verbose:
                print(f"WARNING: Access currently denied to {actual_login}")
            else:
                print("access denied")
        return is_fine
    return asyncio.new_event_loop().run_until_complete(check_leases())


def no_reservation(leases):                             # pylint: disable=w0621
    """
    returns True if nobody currently has a lease
    """
    async def check_leases():
        return not await leases.booked_now_by_anyone()
    return asyncio.new_event_loop().run_until_complete(check_leases())


####################
# NOTE: when adding a new command, please update setup.py as well
supported_subcommands = []                              # pylint: disable=c0103


def subcommand(driver):
    supported_subcommands.append(driver.__name__)
    return driver

####################


@subcommand
def nodes(*argv):
    usage = """
    Just display the hostnames of selected nodes
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    add_selector_arguments(parser)
    args = parser.parse_args(argv)
    selector = selected_selector(args)
    print(" ".join(selector.node_names()))

####################


def cmc_verb(verb, resa_policy, *argv):
    """
    resa_policy can be either
    (*) 'enforce': refuse to send the message if the lease is not there
    (*) 'warn': issue a warning when the lease is not there
    (*) 'none' - or anything else really: does not check the leases
    """
    usage = f"""
    Send verb '{verb}' to the CMC interface of selected nodes"""
    if resa_policy == 'enforce':
        usage += f"\n    {RESERVATION_REQUIRED}"
    config = Config()
    default_timeout = config.value('nodes', 'cmc_default_timeout')

    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-t", "--timeout", action='store',
                        default=default_timeout, type=float,
                        help="Specify global timeout for the whole process")
    add_selector_arguments(parser)
    args = parser.parse_args(argv)

    message_bus = asyncio.Queue()
    leases = Leases(message_bus)                        # pylint: disable=w0621

    if resa_policy in ('warn', 'enforce'):
        reserved = check_reservation(leases, verbose=False)
        if not reserved:
            if resa_policy == 'enforce':
                return 1

    selector = selected_selector(args)
    action = Action(verb, selector)

    return 0 if action.run(message_bus, args.timeout) else 1

#####


@subcommand
def status(*argv):
    return cmc_verb('status', 'warn', *argv)


@subcommand
def on(*argv):                                          # pylint: disable=c0103
    return cmc_verb('on', 'enforce', *argv)


@subcommand
def off(*argv):
    return cmc_verb('off', 'enforce', *argv)


@subcommand
def reset(*argv):
    return cmc_verb('reset', 'enforce', *argv)


@subcommand
def info(*argv):
    return cmc_verb('info', 'warn', *argv)


@subcommand
def usrpstatus(*argv):
    return cmc_verb('usrpstatus', 'warn', *argv)


@subcommand
def usrpon(*argv):
    return cmc_verb('usrpon', 'enforce', *argv)


@subcommand
def usrpoff(*argv):
    return cmc_verb('usrpoff', 'enforce', *argv)


@subcommand
def bothoff(*argv):
    return cmc_verb('bothoff', 'enforce', *argv)


#####
# xxx should be asynchronous
@subcommand
def bye(*argv):
    """
    An alternative implementation of the previous 'all-off' utility
    Switch the lights off when you leave
    """

    usage = """
    Turn off whole testbed
    """
    config = Config()
    # use a dedicated - longer - timeout
    # so be on the safe side
    default_timeout = config.value('nodes', 'cmc_safe_timeout')
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-t", "--timeout", action='store',
                        default=default_timeout, type=float,
                        help="Specify timeout for each phase")
    add_selector_arguments(parser)
    args = parser.parse_args(argv)

    try:
        selector = selected_selector(args, defaults_to_all=True)
        if selector.is_empty():
            selector.use_all_scope()

        message_bus = asyncio.Queue()
        print(f"{20*'='} bothoff {20*'='} (timeout={args.timeout})")
        Action('bothoff', selector).run(message_bus, args.timeout)

        # keep it simple for now
        time.sleep(1)

        # even simpler
        print(f"{20*'='} phones {20*'='}")
        inventory_phones = InventoryPhones()
        for phone in inventory_phones.all_phones():
            command = (f"ssh -i {phone['gw_key']} -o StrictHostKeyChecking=no "
                    f"{phone['gw_user']}@{phone['gw_host']} phone-off")
            print(command)
            os.system(command)

        print(f"{20*'='} pdus {20*'='}")
        inventory_pdus = InventoryPdus.load()
        # this is a working async version
        # but for safety we will do it sequentially
        # because the ssh service in the PDU unit looks fragile
        # async def turn_off_auto_devices():
        #     await asyncio.gather(
        #         *(device.off()
        #                 for device in inventory_pdus.devices
        #                 if device.auto_turn_off))
        # asyncio.run(turn_off_auto_devices())
        for device in inventory_pdus.devices:
            if device.auto_turn_off:
                asyncio.run(device.off())
    except KeyboardInterrupt:
        print("bye: keyboard interrupt - exiting")
        return 1

####################


@subcommand
def load(*argv):
    usage = f"""
    Load an image on selected nodes in parallel
    {RESERVATION_REQUIRED}
    """
    config = Config()
    config.check_binaries()
    imagesrepo = ImagesRepo()

    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-i", "--image", action='store',
                        default=imagesrepo.default(),
                        help="Specify image to load")
    parser.add_argument("-t", "--timeout", action='store',
                        default=config.value('nodes',
                                                 'load_default_timeout'),
                        type=float,
                        help="Specify global timeout for the whole process")
    parser.add_argument("-b", "--bandwidth", action='store',
                        default=config.value('networking', 'bandwidth'),
                        type=int,
                        help="Set bandwidth in Mibps for frisbee uploading")
    parser.add_argument("-c", "--curses", action='store_true', default=False,
                        help="Use curses to provide term-based animation")
    # this is more for debugging
    parser.add_argument("-n", "--no-reset", dest='reset',
                        action='store_false', default=True,
                        help="""use this with nodes that are already
                        running a frisbee image. They won't get reset,
                        neither before or after the frisbee session""")
    add_selector_arguments(parser)
    args = parser.parse_args(argv)

    message_bus = asyncio.Queue()

    selector = selected_selector(args)
    if selector.is_empty():
        parser.print_help()
        return 1
    nodes = [Node(cmc_name, message_bus)                # pylint: disable=w0621
             for cmc_name in selector.cmc_names()]

    # send feedback
    message_bus.put_nowait({'selected_nodes': selector})
    from rhubarbe.logger import logger
    logger.info(f"timeout is {args.timeout}s")
    logger.info(f"bandwidth is {args.bandwidth} Mibps")

    actual_image = imagesrepo.locate_image(args.image, look_in_global=True)
    if not actual_image:
        print(f"Image file {args.image} not found - emergency exit")
        exit(1)

    # send feedback
    message_bus.put_nowait({'loading_image': actual_image})
    display_class = Display if not args.curses else DisplayCurses
    display = display_class(nodes, message_bus)
    loader = ImageLoader(nodes, image=actual_image, bandwidth=args.bandwidth,
                         message_bus=message_bus, display=display)
    return loader.main(reset=args.reset, timeout=args.timeout)

####################


@subcommand
def save(*argv):
    usage = f"""
    Save an image from a node
    Mandatory radical needs to be provided with --output
      This info, together with nodename and date, is stored
      on resulting image in /etc/rhubarbe-image
    {RESERVATION_REQUIRED}
    """

    config = Config()
    config.check_binaries()

    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-o", "--output", action='store', dest='radical',
                        default=None, required=True,
                        help="Mandatory radical to name resulting image")
    parser.add_argument("-t", "--timeout", action='store',
                        default=config.value('nodes',
                                                 'save_default_timeout'),
                        type=float,
                        help="Specify global timeout for the whole process")
    parser.add_argument("-c", "--comment", dest='comment', default=None,
                        help="one-liner comment to insert in "
                        "/etc/rhubarbe-image")
    parser.add_argument("-n", "--no-reset", dest='reset',
                        action='store_false', default=True,
                        help="""use this with a node that is already
                        running a frisbee image. It won't get reset,
                        neither before or after the frisbee session""")
    parser.add_argument("node")
    args = parser.parse_args(argv)

    message_bus = asyncio.Queue()

    selector = Selector()
    selector.add_range(args.node)
    # in case there was one argument but it was not found in inventory
    if len(selector) != 1:
        parser.print_help()
    cmc_name = next(selector.cmc_names())
    node = Node(cmc_name, message_bus)
    nodename = node.control_hostname()

    imagesrepo = ImagesRepo()
    actual_image = imagesrepo.where_to_save(nodename, args.radical)
    message_bus.put_nowait({'info': f"Saving image {actual_image}"})
    # curses has no interest here since we focus on one node
    display_class = Display
    display = display_class([node], message_bus)
    saver = ImageSaver(node, image=actual_image, radical=args.radical,
                       message_bus=message_bus, display=display,
                       comment=args.comment)
    return saver.main(reset=args.reset, timeout=args.timeout)

####################


@subcommand
def wait(*argv):                                        # pylint: disable=r0914
    usage = """
    Wait for selected nodes to be reachable by ssh
    Returns 0 if all nodes indeed are reachable
    """
    # suppress info log messages from asyncssh
    asyncssh_set_log_level(logging.WARNING)

    config = Config()
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-c", "--curses", action='store_true', default=False,
                        help="Use curses to provide term-based animation")
    parser.add_argument("-t", "--timeout", action='store',
                        default=config.value('nodes', 'wait_default_timeout'),
                        type=float,
                        help="Specify global timeout for the whole process")
    parser.add_argument("-b", "--backoff", action='store',
                        default=config.value('networking', 'ssh_backoff'),
                        type=float,
                        help="Specify backoff average between "
                        "attempts to ssh connect")
    parser.add_argument("-u", "--user", default="root",
                        help="select other username")
    # really dont' write anything
    parser.add_argument("-s", "--silent", action='store_true', default=False)
    parser.add_argument("-v", "--verbose", action='store_true', default=False)

    add_selector_arguments(parser)
    args = parser.parse_args(argv)

    # --curses implies --verbose otherwise nothing shows up
    if args.curses:
        args.verbose = True

    selector = selected_selector(args)
    message_bus = asyncio.Queue()

    if args.verbose:
        message_bus.put_nowait({'selected_nodes': selector})
    from rhubarbe.logger import logger
    logger.info(f"wait: backoff is {args.backoff} "
                f"and global timeout is {args.timeout}")

    nodes = [Node(cmc_name, message_bus)                # pylint: disable=w0621
             for cmc_name in selector.cmc_names()]
    sshs = [SshProxy(node, username=args.user, verbose=args.verbose)
            for node in nodes]
    jobs = [Job(ssh.wait_for(args.backoff), critical=True) for ssh in sshs]

    display_class = Display if not args.curses else DisplayCurses
    display = display_class(nodes, message_bus)

    # have the display class run forever until the other ones are done
    scheduler = Scheduler(Job(display.run(), forever=True, critical=True),
                          *jobs,
                          timeout=args.timeout,
                          critical=False)
    try:
        orchestration = scheduler.run()
        if orchestration:
            return 0
        else:
            if args.verbose:
                scheduler.debrief(silence_done_jobs=True)
            return 1
    except KeyboardInterrupt:
        print("rhubarbe-wait : keyboard interrupt - exiting")
        # xxx
        return 1
    finally:
        display.epilogue()
        if not args.silent:
            for ssh in sshs:
                print(f"{ssh.node}:ssh {'OK' if ssh.status else 'KO'}")

####################


@subcommand
def images(*argv):
    usage = """
    Display available images
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-l", "--labeled-only", dest="labeled",
                        action='store_true', default=False,
                        help="""if specified, only images that have at least
                        one symlink to them are shown""")
    parser.add_argument("-p", "--public-only",
                        action="store_true", default=False,
                        help="displays only publicly visible images")
    parser.add_argument("-s", "--size", dest='sort_size',
                        action='store_true', default=None,
                        help="sort by size (default is by name)")
    parser.add_argument("-d", "--date", dest='sort_date',
                        action='store_true', default=None,
                        help="sort by date")
    parser.add_argument("-r", "--reverse",
                        action='store_true', default=False,
                        help="reverse sort")
    parser.add_argument("-n", "--narrow",
                        action='store_true', default=False,
                        help="""default is to show full paths, with this option
                        only radicals are displayed""")
    parser.add_argument("focus", nargs="*", type=str,
                        help="if provided, only images that contain "
                        "one of these strings are displayed")
    args = parser.parse_args(argv)
    imagesrepo = ImagesRepo()
    if args.sort_size is not None:
        args.sort_by = 'size'
    elif args.sort_date is not None:
        args.sort_by = 'date'
    else:
        args.sort_by = 'name'

    # if focus is an empty list, then everything is shown
    return imagesrepo.images(
        args.focus, args.sort_by, args.reverse,
        args.labeled, args.public_only, args.narrow)

####################


@subcommand
def resolve(*argv):
    usage = """for each input, find out and display
    what file exactly would be used if used with load
    and possible siblings if verbose mode is selected
    """

    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
#    parser.add_argument("-r", "--reverse", action='store_true', default=False,
#                        help="reverse sort")
    parser.add_argument("-v", "--verbose", action='store_true', default=False,
                        help="show all files (i.e. with all symlinks)")
    parser.add_argument("focus", type=str,
                        help="the image radical name to resolve")
    args = parser.parse_args(argv)
    imagesrepo = ImagesRepo()
    # if focus is an empty list, then everything is shown
    return imagesrepo.resolve(args.focus, args.verbose)

####################


@subcommand
def share(*argv):
    usage = """
    Install privately-stored images into the global images repo
    Destination name is derived from the radical provided at save-time
      i.e. trailing part after saving__<date>__

    With the -a option, it is possible to make the new image
    **also** visible under that alias name.

    If your account is enabled in /etc/sudoers.d/rhubarbe-share,
    the command will actually perform the mv operation
    Requires to be run through 'sudo rhubarbe-share'
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-a", "--alias-name", dest='alias',
                        action='store', default=None,
                        help="also create a symlink of that name")
    # default=None so that imagesrepo.share can compute a default
    parser.add_argument("-n", "--dry-run",
                        default=None, action='store_true',
                        help="Only show what would be done "
                        "(default unless running under sudo")
    parser.add_argument("-f", "--force",
                        default=False, action='store_true',
                        help="Will move files even if destination exists")
    parser.add_argument("-c", "--clean",
                        default=False, action='store_true',
                        help="""Will remove other matches than the one that
                        gets promoted. In other words, useful when one name
                        has several matches and only the last one is desired.
                        Typically after a successful rsave""")
    parser.add_argument("image", type=str)
    args = parser.parse_args(argv)

    imagesrepo = ImagesRepo()
    return imagesrepo.share(
        args.image, args.alias, args.dry_run, args.force, args.clean)

####################


@subcommand
def leases(*argv):
    usage = """
    Unless otherwise specified, displays current leases, from today onwards
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument('-c', '--check', action='store_true', default=False,
                        help="Check if you currently have a lease")
    parser.add_argument('-i', '--interactive', action='store_true',
                        default=False,
                        help="Interactively prompt for commands "
                             "(create, update, delete)")
    args = parser.parse_args(argv)

    message_bus = asyncio.Queue()
    leases = Leases(message_bus)
    if args.check:
        access = check_reservation(leases, verbose=True)
        return 0 if access else 1
    else:
        loop = asyncio.new_event_loop()
        loop.run_until_complete(leases.main(args.interactive))
        loop.close()
        return 0

####################


@subcommand
def book(*argv):
    exit(0 if Book.main(argv) else 1)

####################
from rhubarbe.logger import monitor_logger
from rhubarbe.logger import r2lab_sidecar_logger

def set_loggers_level(verbose, debug):
    print("setting logger level on r2lab-sidecar and monitor")
    if debug:
        monitor_logger.setLevel(logging.DEBUG)
        r2lab_sidecar_logger.setLevel('DEBUG')
    elif verbose:
        monitor_logger.setLevel('INFO')
        r2lab_sidecar_logger.setLevel('INFO')
    else:
        monitor_logger.setLevel('WARNING')
        r2lab_sidecar_logger.setLevel('WARNING')



@subcommand
def monitornodes(*argv):                                # pylint: disable=r0914

    usage = """
    Cyclic probe all selected nodes, and reports
    real-time status at a sidecar service over websockets
    """
    config = Config()
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '-c', "--cycle",
        default=config.value('monitor', 'cycle_nodes'),
        type=float,
        help="Delay to wait between 2 probes of each node")
    parser.add_argument(
        "-u", "--sidecar-url", dest="sidecar_url",
        default=Config().value('sidecar', 'url'),
        help="url for the sidecar server")
    parser.add_argument(
        "-v", "--verbose", action='store_true')
    parser.add_argument(
        "-d", "--debug", action='store_true')
    add_selector_arguments(parser)
    args = parser.parse_args(argv)

    (debug, verbose) = (args.debug, args.verbose)
    set_loggers_level(verbose, debug)

    selector = selected_selector(args)
    message_bus = asyncio.Queue()

    # xxx having to feed a Display instance with nodes
    # at creation time is a nuisance
    display = Display([], message_bus)

    monitor_logger.info({'selected_nodes': selector})
    monitornodes = MonitorNodes(selector.cmc_names(),
                                message_bus=message_bus,
                                cycle=args.cycle,
                                sidecar_url=args.sidecar_url,
                                verbose=args.verbose)

    async def async_main():
        # run both the core and the log loop in parallel
        await asyncio.gather(monitornodes.run_forever(),
                             display.run())

    MonitorLoop("monitornodes").run(async_main())
    return 0

####################


@subcommand
def monitorphones(*argv):

    usage = """
    Cyclic probe all known phones, and reports real-time status
    at a sidecar service over websockets
    """
    config = Config()
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-c", "--cycle",
        default=config.value('monitor', 'cycle_phones'),
        type=float,
        help="Delay to wait between 2 probes of each phone")
    parser.add_argument(
        "-u", "--sidecar-url", dest="sidecar_url",
        default=Config().value('sidecar', 'url'),
        help="url for the sidecar server")
    parser.add_argument(
        "-v", "--verbose", action='store_true')
    parser.add_argument(
        "-d", "--debug", action='store_true')
    args = parser.parse_args(argv)

    (debug, verbose) = (args.debug, args.verbose)
    del args.debug
    set_loggers_level(verbose, debug)

    monitor_logger.info("Using all phones")
    monitorphones = MonitorPhones(**vars(args))

    MonitorLoop("monitorphones").run(monitorphones.run_forever())
    return 0

####################


@subcommand
def monitorpdus(*argv):

    usage = """
    Cyclic probe all known pdus, and reports real-time status
    at a sidecar service over websockets
    """
    config = Config()
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-c", "--cycle",
        default=config.value('monitor', 'cycle_pdus'),
        type=float,
        help="Delay to wait between 2 probes of each pdu")
    parser.add_argument(
        "-u", "--sidecar-url", dest="sidecar_url",
        default=Config().value('sidecar', 'url'),
        help="url for the sidecar server")
    parser.add_argument(
        "-v", "--verbose", action='store_true')
    parser.add_argument(
        "-d", "--debug", action='store_true')
    parser.add_argument("names", nargs='*', help="optionally provide device names")
    args = parser.parse_args(argv)

    (debug, verbose) = (args.debug, args.verbose)
    set_loggers_level(verbose, debug)
    # not supported by MonitorPdus
    del args.debug


    monitor_logger.info(f"Monitoring pdus: {args.names or 'all'}")
    monitorpdus = MonitorPdus(**vars(args))

    MonitorLoop("monitorpdus").run(monitorpdus.run_forever())

    return 0

####################


@subcommand
def monitorleases(*argv):

    from rhubarbe.logger import monitor_logger as logger

    usage="""
    Cyclic check of leases; also reacts to 'request' messages on
    the sidecar channel, which triggers leases acquisition right away.
    See config for defaults.
    """
    #config = Config()
    parser = ArgumentParser(
        usage=usage, formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-u", "--sidecar-url", dest="sidecar_url",
        default=Config().value('sidecar', 'url'),
        help="url for the sidecar server")
    parser.add_argument(
        "-v", "--verbose", default=False, action='store_true')
    args = parser.parse_args(argv)

    logger.info("monitoring leases")

    message_bus = asyncio.Queue()
    monitorleases = MonitorLeases(
        message_bus, args.sidecar_url, args.verbose)

    MonitorLoop("monitorleases").run(monitorleases.run_forever())

    return 0


####################


@subcommand
def accountsmanager(*argv):

    usage = "The core of the accounts manager; reserved to root"

    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("-c", "--cycle",
                        help="Set cycle in seconds; 0 means run only once;"
                             " default from config file.",
                        default=None)
    args = parser.parse_args(argv)

    accounts_manager = AccountsManager()
    return accounts_manager.main(args.cycle)

####################


@subcommand
def inventory(*argv):
    usage = """
    Display inventory
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.parse_args(argv)
    inventory = InventoryNodes()
    inventory.display(verbose=True)
    return 0

####################


@subcommand
def config(*argv):
    usage = """
    Display global configuration
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument("sections", nargs='*',
                        type=str,
                        help="config section(s) to display")
    args = parser.parse_args(argv)
    config = Config()
    config.display(args.sections)
    return 0

####################


@subcommand
def template(*argv):
    usage = """
    Show the template file to create the nodes and phones inventory
    under /etc/rhubarbe
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        "-n", "--nodes", dest='nodes',
        action='store_true', default=False,
        help="Show template for /etc/rhubarbe/inventory-nodes.json")
    parser.add_argument(
        "-p", "--phones", dest='phones',
        action='store_true', default=False,
        help="Show template for /etc/rhubarbe/inventory-phones.json")
    args = parser.parse_args(argv)

    def show_template(nodes_or_phones):
        template = resource_string(
            'rhubarbe', f"config/inventory-{nodes_or_phones}.json.template")
        print("# =========="
              f" template for /etc/rhubarbe/inventory-{nodes_or_phones}.json")
        print(template.decode(encoding='utf-8'))

    # pick -n by default
    if not args.nodes and not args.phones:
        args.nodes = True

    for nodes_or_phones in ('nodes', 'phones'):
        selected = getattr(args, nodes_or_phones)
        if selected:
            show_template(nodes_or_phones)
    return 0

####################


@subcommand
def pdu(*argv):
    usage = """
    manage PDUs; examples:

        rhubarbe-pdu list               # lists the known PDUs and devices by name
                                        # all forms of list show STATIC INFO only
                                        # (no remote device involved)
        rhubarbe-pdu list r2lab         # here r2lab is the name of a pdu_host
                                        # here again it's a static info
        rhubarbe-pdu list jaguar        # works with devices too


        rhubarbe-pdu status r2lab       # this time the PDU is probed
                                        # for its current status
        rhubarbe-pdu status jaguar      # works with devices too


        rhubarbe-pdu on jaguar
        rhubarbe-pdu off panther
        rhubarbe-pdu reset n300

    Note that for the OFF and RESET variants, there is an attempt to perform
    a soft shutdown first (i.e. to contact the device by ssh and to run
    `shutdown -h now` in that context)
    for that the device needs to be properly configured
    """
    parser = ArgumentParser(usage=usage,
                            formatter_class=ArgumentDefaultsHelpFormatter)
    # not yet used
    parser.add_argument("-v", "--verbose", default=False, action='store_true')
    parser.add_argument("command", help="one of the supported contrib commands")
    parser.add_argument("extras", nargs="*", help="additional args are passed as is")
    args = parser.parse_args(argv)

    if args.verbose:
        import rhubarbe.inventorypdus
        rhubarbe.inventorypdus.VERBOSE = True

    command, extras = args.command, args.extras

    def die():
        parser.print_help()
        exit(255)

    # export env variables picked in config
    # and uppercased as is the tradition for global env. variables

    try:
        inventory_pdus = InventoryPdus.load()   # pylint: disable=no-member

        match command:
            case 'list':
                inventory_pdus.list(extras)

            case 'status':
                if not extras:
                    inventory_pdus.status()
                else:
                    for name in extras:
                        try:
                            pdu_host = inventory_pdus.get_pdu_host(name)
                            retcod = asyncio.run(pdu_host.probe())
                            continue
                        except ValueError:
                            pass
                        try:
                            device = inventory_pdus.get_device(name)
                            retcod = asyncio.run(device.run_action('status'))
                            continue
                        except ValueError:
                            print(f"unknown device {name}")
                exit(0)
            case 'on' | 'off' | 'reset':
                match extras:
                    case (name, ):
                        # status accepts a pdu_host name
                        if command == 'status':
                            try:
                                pdu_host = inventory_pdus.get_pdu_host(name)
                                retcod = asyncio.run(pdu_host.probe())
                                exit(retcod)
                            except ValueError:
                                pass
                        # otherwise a device name is expected
                        try:
                            device = inventory_pdus.get_device(name)
                            retcod = asyncio.run(device.run_action(command))
                            exit(retcod)
                        except ValueError as exc:
                            print(exc)
                            exit(255)
                    case _:
                        print(f"missing device name for {command}")
                        die()
            case _:
                print(f"unknown command {command}")
                die()
    except ValueError as exc:
        print(exc)
        exit(255)
####################


@subcommand
def version(*_):
    from rhubarbe.version import __version__
    print(f"rhubarbe version {__version__}")
    return 0



"""
Utility for loading the config
as a layered cake based on several locations
"""

import os
import socket
import configparser
from pathlib import Path

from pkg_resources import resource_exists, resource_filename

from rhubarbe.singleton import Singleton
from rhubarbe.logger import logger

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111,w1202,r1705
# pylint: disable=logging-fstring-interpolation, import-outside-toplevel

# http://setuptools.readthedocs.io/en/latest/pkg_resources.html
# says to not use os.path to check for resources
DEFAULT_EXISTS = resource_exists('rhubarbe', 'config/rhubarbe.conf')
DEFAULT_LOCATION = resource_filename('rhubarbe', 'config/rhubarbe.conf')

LOCATIONS = [
    # all the files found in these locations are considered
    # and are all loaded in this order, so the last ones
    # overwrite the first ones
    # path, mandatory
    (DEFAULT_LOCATION, DEFAULT_EXISTS, True),
    ("/etc/rhubarbe/rhubarbe.conf.local", None, False),
    (str(Path.home() / ".rhubarbe.conf"), None, False),
    ("./rhubarbe.conf", None, False),
    ("./rhubarbe.conf.local", None, False),
]


class ConfigException(Exception):
    pass


class Config(metaclass=Singleton):

    def __init__(self):
        self.parser = configparser.ConfigParser()
        self.files = []
        # load all configurations when they exist
        for location, exists, mandatory in LOCATIONS:
            if exists is None:
                exists = Path(location).exists()
            if exists:
                self.files.append(location)
                self.parser.read(location)
                logger.info(f"Loaded config from {location}")
            elif mandatory:
                raise ConfigException(
                    f"Missing mandatory config file {location}")
        #
        self._hostname = None

    def local_hostname(self):
        if not self._hostname:
            self._hostname = socket.gethostname().split('.')[0]
        return self._hostname

    @staticmethod
    def get_or_raise(dictobj, section, key):
        res = dictobj.get(key, None)
        if res is not None:
            return res
        else:
            raise ConfigException(
                f"rhubarbe config: missing entry section={section} key={key}")

    def value(self, section, flag):
        if section not in self.parser:
            raise ConfigException(f"No such section {section} in config")
        config_section = self.parser[section]
        hostname = self.local_hostname()
        key = f"{flag}.{hostname}"
        return config_section.get(key, None) \
            or self.get_or_raise(config_section, section, flag)

    def full_section(self, section) -> dict:
        if section not in self.parser:
            raise ConfigException(f"No such section {section} in config")
        config_section = self.parser[section]
        return {key: config_section[key] for key in config_section.keys()}


    # for now
    # the foreseeable tricky part is, this should be a coroutine..
    def available_frisbee_port(self):
        return self.value('networking', 'port')

    # maybe this one too
    def local_control_ip(self):
        # if specified in the config file, then use that
        if ('networking' in self.parser and
                'local_control_ip' in self.parser['networking']):
            return self.parser['networking']['local_control_ip']
        # but otherwise guess it
        # do not import at toplevel to avoid import loop
        from rhubarbe.inventorynodes import InventoryNodes
        the_inventory = InventoryNodes()
        from rhubarbe.guessip import local_ip_on_same_network_as
        ipaddr, _ = local_ip_on_same_network_as(
            the_inventory.one_control_interface())
        return ipaddr

    def display(self, sections):
        for i, file in enumerate(self.files):
            print(f"{i+1}-th config file = {file}")

        def match(section, sections):
            return not sections or section in sections
        for sname, section in sorted(self.parser.items()):
            if match(sname, sections) and section:
                print(10*'=', f" section {sname}")
                for fname, value in sorted(section.items()):
                    print(f"{fname} = {value}")

    @staticmethod
    def check_file_in_path(binary):
        PATH = os.environ['PATH']                       # pylint: disable=c0103
        paths = ['/'] + [p for p in PATH.split(':') if p]
        for path in paths:
            full = Path(path) / binary
            if full.exists():
                return Path(path) / binary
        return False

    def check_binaries(self):
        # imagezip and frisbee are required on the pxe image only
        names = ('server', 'netcat')
        binaries = [self.value('frisbee', name) for name in names]

        for binary in binaries:
            checked = self.check_file_in_path(binary)
            if not checked:
                message = f"Binary {binary} not found in PATH"
                logger.critical(message)
                raise Exception(message)
            else:
                print(f"Found binary {binary} as {checked}")



"""
create/configure logger objects
"""

# c0111 no docstrings yet
# c0103 constants names should be UPPER_CASE
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# r0903 too few public methods
# pylint: disable=c0103, w0703

import sys
from pathlib import Path

import logging
import logging.config
from pathlib import Path

# with systemd it's sooo simpler to log on stdout, that gets managed by journal
# so, we essentially need
# * one all-purpose handler that goes into $HOME/rhubarbe.log
# * one special handler for monitor*s that goes onto stdout -> journal
# * one special handler for accounts - ditto but with a shorter layout

rhubarbe_logging_config = {
    'version': 1,
    'disable_existing_loggers': True,
    'formatters': {
        'standard': {
            'format': '%(asctime)s %(levelname)s '
                      '%(filename)s:%(lineno)d %(message)s',
            'datefmt': '%m-%d %H:%M:%S'
        },
        'shorter': {
            'format': '%(asctime)s %(levelname)s %(message)s',
            'datefmt': '%d %H:%M:%S'
        },
    },
    'handlers': {
        # set to DEBUG, so one can tweak the loggers later on
        'rhubarbe': {
            'level': 'DEBUG',
            'class': 'logging.FileHandler',
            'formatter': 'standard',
            'filename': f'{str(Path.home())}/rhubarbe.log',
        },
        'monitor': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'standard',
            'stream': sys.stdout,
        },
        'accounts': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'shorter',
            'stream': sys.stdout,
        },
    },
    'loggers': {
        # the default is INFO and the code sometimes set this to DEBUG
        'monitor': {
            'handlers': ['monitor'],
            'level': 'DEBUG',
            'propagate': False,
        },
        # this one is for the r2lab-python library
        # that is used in the monitors
        'r2lab-sidecar': {
            'handlers': ['monitor'],
            'level': 'DEBUG',
            'propagate': False,
        },
        'accounts': {
            'handlers': ['accounts'],
            'level': 'INFO',
            'propagate': False,
        },
        'rhubarbe': {
            'handlers': ['rhubarbe'],
            'level': 'INFO',
            'propagate': False,
        },
    },
}

logging.config.dictConfig(rhubarbe_logging_config)

# general case:
logger = logging.getLogger('rhubarbe')

# monitor
monitor_logger = logging.getLogger('monitor')
r2lab_sidecar_logger = logging.getLogger('r2lab-sidecar')


# accounts
accounts_logger = logging.getLogger('accounts')

####################
# test
if __name__ == '__main__':
    logger.info("in rhubarbe")
    monitor_logger.info("in monitor")



#!/usr/bin/env python3
MAX_BUF = 4096

import asyncio, telnetlib3

print(telnetlib3.__file__)

print(telnetlib3.__get_version())

async def shell(reader, writer):

    recv = await reader.read(MAX_BUF)
    print(f'--init--(({recv}))', end='', flush=True)

    commands = []
    commands.append("echo 1")
    commands.append("echo 2; echo 3")
    commands.append("echo 4; echo 5; echo 6;")
    commands.append("""[ -d /mnt ] || mkdir /mnt; mount /dev/sda1 /mnt && { echo "2019-05-29@14:44 - node fit37  - image f-strings - by root" >> /mnt/etc/rhubarbe-image ; umount /mnt; } ; imagezip -o -z1 /dev/sda - | nc 192.168.3.100 10001; echo STATUS=$?""")
    commands.append('exit')

    for command in commands:
        print(f'[[{command}]]')
        writer.write(command + '\n')
        recv = await reader.read(MAX_BUF)
        if reader.at_eof():
            # End of File
            break
        print(f"(({recv}))", end="", flush=True)
    while not reader.at_eof():
        recv = await reader.read(MAX_BUF)
        print(f"::{recv}::", end="", flush=True)

loop = asyncio.new_event_loop()
coro = telnetlib3.open_connection('fit04', 23, shell=shell)
reader, writer = loop.run_until_complete(coro)
loop.run_until_complete(writer.protocol.waiter_closed)



"""
Parsing the output of the remote frisbee app
"""

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, w0201, r1705, w1201, w1202, w1203

import re

from rhubarbe.logger import logger
from rhubarbe.telnet import TelnetProxy
from rhubarbe.config import Config


class FrisbeeParser:
    def __init__(self, proxy):
        self.proxy = proxy
        self.total_chunks = 0

    def ip(self):
        return self.proxy.control_ip

    def feedback(self, field, msg):
        self.proxy.message_bus.put_nowait({'ip': self.ip(), field: msg})

    def send_percent(self, percent):
        self.feedback('percent', percent)

    # parse frisbee output
    # tentatively ported from nitos_testbed ruby code but never tested
    matcher_new_style_progress = \
        re.compile(r'[\.sz]{60,75}\s+\d+\s+(?P<remaining_chunks>\d+)')
    matcher_total_chunks = \
        re.compile(r'.*team after (?P<time>[0-9\.]*).*'
                   r'File is (?P<total_chunks>[0-9]+) chunks.*')
    matcher_old_style_progress = \
        re.compile(r'^Progress:\s+(?P<percent>[\d]+)%.*')
    matcher_final_report = \
        re.compile(r'^Wrote\s+(?P<total>\d+)\s+bytes \((?P<actual>\d+).*')
    matcher_short_write = \
        re.compile(r'.*Short write.*')
    matcher_status = \
        re.compile(r'FRISBEE-STATUS=(?P<status>\d+)')

    def parse_line(self, line):
        #
        match = self.matcher_new_style_progress.match(line)
        if match:
            if self.total_chunks == 0:
                logger.error(
                    "ip={self.ip()}: new frisbee: cannot report progress, "
                    "missing total chunks")
                return
            percent = int(100 * (1 - int(match.group('remaining_chunks'))
                                 / self.total_chunks))
            self.send_percent(percent)
        #
        match = self.matcher_total_chunks.match(line)
        if match:
            self.total_chunks = int(match.group('total_chunks'))
            self.send_percent(0)
            return
        #
        match = self.matcher_old_style_progress.match(line)
        if match:
            self.send_percent(match.group('percent'))
            return
        #
        match = self.matcher_final_report.match(line)
        if match:
            logger.info(f"ip={self.ip()} FRISBEE END: "
                        f"total = {match.group('total')} bytes, "
                        f"actual = {match.group('actual')} bytes")
            self.send_percent(100)
            return
        #
        match = self.matcher_short_write.match(line)
        if match:
            self.feedback('frisbee_error',
                          "Something went wrong with frisbee (short write...)")
            return
        #
        match = self.matcher_status.match(line)
        if match:
            status = int(match.group('status'))
            self.feedback('frisbee_retcod', status)


class Frisbee(TelnetProxy):

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)
        self.parser = FrisbeeParser(self)

    def line_callback(self, line):
        self.parser.parse_line(line)

    async def run(self, multicast_ip, port):
        control_ip = self.control_ip
        the_config = Config()
        client = the_config.value('frisbee', 'client')
        hdd = the_config.value('frisbee', 'hard_drive')
        self.command = (
            f"{client} -i {control_ip} -m {multicast_ip} -p {port} {hdd}")

        logger.info(f"on {self.control_ip} : running command {self.command}")
        await self.feedback('frisbee_status', "starting frisbee client")

        retcod = await self.session([self.command])

        logger.info(f"frisbee on {self.control_ip} returned {retcod}")

        return retcod



"""
TelnetProxy is what controls the telnet connection to the node,
it is subclassed as Frisbee and ImageZip

As per the design of telnetlib3, it uses a couple helper classes
* the shell (telnetlib3.TerminalShell) is what receives the session's output,
  so we have for example a FrisbeeParser class that acts as such a shell
* a client (telnetlib3.TelnetClient);
  that we specialize so we can propagate our own stuff
  (the telnetproxy instance primarily) down to FrisbeeParser

This class essentially is the mother class for Frisbee and ImageZip
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, w0703, w1202

import random
import asyncio
import telnetlib3

from rhubarbe.logger import logger
from rhubarbe.config import Config

MAX_BUF = 16 * 1024


class TelnetClient(telnetlib3.TelnetClient):
    """
    this specialization of TelnetClient is meant for FrisbeeParser
    to retrieve its correponding TelnetProxy instance
    """
    def __init__(self, proxy, *args, **kwds):
        self.proxy = proxy
        super().__init__(*args, **kwds)



class TelnetProxy:
    """
    a convenience class that help us
    * wait for the telnet server to come up
    * invoke frisbee when rload'ing
    * invoke imagezip when rsave'ing
    """

    def __init__(self, control_ip, message_bus):
        self.control_ip = control_ip
        self.message_bus = message_bus
        # config
        the_config = Config()
        self.port = int(the_config.value('networking', 'telnet_port'))
        self.backoff = float(the_config.value('networking', 'telnet_backoff'))
        self.connect_timeout = float(the_config.value('networking', 'telnet_timeout'))
        self.connect_minwait = float(the_config.value('networking', 'telnet_connect_minwait'))
        self.connect_maxwait = float(the_config.value('networking', 'telnet_connect_maxwait'))
        # internals
        self.running = False
        self._reader = None
        self._writer = None


    def is_ready(self):
        return self._writer is not None


    async def feedback(self, field, msg):
        await self.message_bus.put({'ip': self.control_ip, field: msg})


    async def try_to_connect(self):

        # a little closure to capture our ip and expose it to the parser
        def client_factory():
            return TelnetClient(proxy=self, encoding='utf-8')

        await self.feedback('frisbee_status', "trying to telnet..")
        logger.info(f"Trying to telnet on {self.control_ip}")
        try:
            self._reader, self._writer = await asyncio.wait_for(
                telnetlib3.open_connection(
                    self.control_ip, 23, shell=None,
                    connect_minwait=self.connect_minwait,
                    connect_maxwait=self.connect_maxwait),
                timeout = self.connect_timeout)
        except (asyncio.TimeoutError, OSError) as exc:
            self._reader, self._writer = None, None
        except Exception as exc:
            logger.exception(f"telnet connect: unexpected exception {exc}")


    async def wait_until_connect(self):
        """
        wait for the telnet server to come up
        this has no native timeout mechanism
        """
        while True:
            await self.try_to_connect()
            if self.is_ready():
                return True
            else:
                backoff = self.backoff*(0.5 + random.random())
                await self.feedback('frisbee_status',
                                    f"backing off for {backoff:.3}s")
                await asyncio.sleep(backoff)


    def line_callback(self, line):
        """
        this is intended to be redefined by daughter classes
        it will be called with each piece of data that comes back
        as a result of invoking session()
        no line-asembling is done in the present class for now,
        returned input is triggered as it comes
        """
        logger.error(f"redefine telnet.line_callback()")


    async def session(self, commands):
        """
        given a list of shell commands, will issue them
        before exiting
        all the commands are fired at once in sequence

        return is a boolean that says whether the last command ran OK
        i.e. return is True if retcod is 0, False otherwise
        """

        commands.append('echo _TELNET_STATUS=$?')
        commands.append('exit')

        def parse_status(line):
            os_status = int(line.strip().replace("_TELNET_STATUS=", ""))
            return os_status == 0

        for command in commands:
            logger.debug(f"telnet -> {command}")
            # print(f'[[{command}]]')
            self._writer.write(command + '\n')


        self.running = True
        retcod = False

        line = ""
        while True:
            if self._reader.at_eof():
                break
            recv = await self._reader.read(MAX_BUF)
            for incoming in recv:
                if incoming == "\n":
                    logger.debug(f"telnet <- {line}")
                    if line.startswith("_TELNET_STATUS"):
                        retcod = parse_status(line)
                    self.line_callback(line)
                    line = ""
                else:
                    line += incoming

        self.running = False

        return retcod



"""
The contents of one directory that holds images
"""

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, r1705

import sys
import os
import time
import re
import glob
from pathlib import Path
#from itertools import chain
from collections import defaultdict

from typing import Iterator, List

from rhubarbe.config import Config
from rhubarbe.singleton import Singleton

# to indicate that 0 is OK and others are KO
OsRetcod = int

DEBUG = False

# use __ instead of == because = ruins bash completion
SUFFIX = ".ndz"
SAVING = 'saving'
SEP = '__'
TIME_FORMAT = "%Y-%m-%d@%H-%M"

SEP_RE_PATTERN = "[-_=]{2}"
DATE_RE_PATTERN = "[0-9]{4}-[0-9]{2}-[0-9]{2}@[0-9]{2}.[0-9]{2}"
# this will be fed into .format() - needs double {{ and }}
HOSTNAME_RE_PATTERN_F = "{regularname}[0-9]{{2}}"

SEP_GLOB_PATTERN = "[-_=]" * 2
DATE_GLOB_PATTERN = "[0-9-@]" * 16
HOSTNAME_GLOB_PATTERN_F = "{regularname}[0-9][0-9]"


def root_privileges():
    # surprisingly, even when running under sudo we have all 3 set to 0
    real, effective, saved = os.getresuid()         # pylint: disable=e1101
    is_root = (real == 0) and (effective == 0) and (saved == 0)
    return is_root


class ImagePath:                                 # pylint: disable=r0902, r0903
    def __init__(self, repo, path):
        self.repo = repo
        self.path = Path(path)
        # pylint: disable=w0212
        self.radical = self.repo._radical_part(self.path)
        self.stem = self.path.stem
        self.is_official = self.radical == self.stem
        # just in case
        self.readable = None
        self._infos()

    def _infos(self):
        try:
            with self.path.open():
                pass
            self.readable = True
        except OSError:
            print(f"WARNING unreadable path {self}")
            self.readable = False
            self.mtime = 0
            self.size = 0
            return
        stat = self.path.stat()
        self.mtime = stat.st_mtime
        self.size = stat.st_size
        self.inode = stat.st_ino
        self.is_alias = self.path.is_symlink()

    def __str__(self):
        return str(self.path)

    def __repr__(self):
        return self._to_display(show_path=True)

    def __eq__(self, other):
        return self.path == other.path

    def _to_display(self, show_path, radical_width=None, prefix=''):
        # date/time on 10+1+5 = 16
        # 1 sep
        # size on 8
        # 2 sep
        # radical on 30 : 56
        # 2 sep
        radical_width = (radical_width if radical_width is not None
                         else len(self.radical))
        result = prefix
        date = time.strftime("%Y-%m-%d %H:%M", time.localtime(self.mtime))
        if not self.is_alias:
            result += f"{date:<16s} "
            result += f"{self.bytes2human(self.size):>8s}"
        else:
            result += f"{'':<16s} "
            result += f"{'aka':>8s}"
        result += f"  {self.radical:{radical_width}}"
        if show_path and not self.is_alias:
            result += f"  {self.path}"
        return result

    @staticmethod
    def bytes2human(size, repr_format="{value:.2f}{symbol}"):
        """
        >>> bytes2human(10000)
        '9K'
        >>> bytes2human(100001221)
        '95M'
        """
        symbols = ('B', 'K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y')
        prefix = {}
        for i, symbol in enumerate(symbols[1:]):
            prefix[symbol] = 1 << (i+1)*10
        for symbol in reversed(symbols[1:]):
            if size >= prefix[symbol]:
                value = float(size) / prefix[symbol]
                return repr_format.format(value=value, symbol=symbol)
        return repr_format.format(symbol=symbols[0], value=size)


class ImageCluster:
    """
    a cluster is a set of ImagePath that share the same actual file
    typically it is one plain file and n-1 symlinks
    """
    def __init__(self, repo, image_paths):
        self.repo = repo
        self.image_paths = image_paths
        inodes = {path.inode for path in self.image_paths}
        if len(inodes) != 1:
            print("OOPS, wrong clustering!")
        # the image_path that is no symlink/alias
        self._regular = None
        self.aliases = 0
        for path in self.image_paths:
            if not path.is_alias:
                self._regular = path
            else:
                self.aliases += 1
        if not self._regular:
            print(f"OOPS - no regular file found in cluster "
                  f"with {self.image_paths[0]} that has {self.aliases} aliases")
        self.image_paths.sort(
            key=lambda image_path: image_path.mtime, reverse=True)


    def __iter__(self):
        return iter(self.image_paths)

    def __repr__(self):
        return repr(self.regular)

    @property
    def regular(self):
        return self._regular or self.image_paths[0]


    def _display(self, long_format, radical_width):
        # pylint: disable=w0212
        print(self.regular._to_display(long_format, radical_width))
        for image_path in self.image_paths:
            if image_path == self.regular:
                continue
            print(image_path._to_display(long_format, radical_width))

    def _max_radical(self):
        return max(len(image.radical) for image in self)


class ImagesRepo(metaclass=Singleton):
    def __init__(self):
        the_config = Config()
        self.public = Path(the_config.value('frisbee', 'images_dir'))
        self.default_name = the_config.value('frisbee', 'default_image')
        # for building hostname patterns
        regularname = the_config.value('testbed', 'regularname')
        self._radical_re_matcher = re.compile(
            SEP.join([
                f"{SAVING}",
                f"{regularname}[0-9][0-9]",
                f"{DATE_RE_PATTERN}",
                f"(?P<radical>.+)",
                ]))


    def default(self) -> str:
        return str(self.public / self.default_name)

    @staticmethod
    def where_to_save(nodename, radical) -> str:
        """
        given a nodename, plus a user-provided image name
        computes the actual path where to store an image
        as of 3.0.7, root is deemed to not be likely to build images
        so no more special case for uid==0
        """
        stem = SEP.join([SAVING, nodename,
                         time.strftime(TIME_FORMAT), radical])
        return stem + SUFFIX

    def _radical_part(self, path):
        """
        incoming can be a filename, possibly without an extension
        the saving__ blabla is extracted and only the part
        given at save-time is returned
        it is kind of the reverse of where_to_save
        """
        stem = Path(path).stem
        match = self._radical_re_matcher.match(stem)
        if match:
            return match.group('radical')
        return stem

    def _iterate_images(self, directory, predicate) -> Iterator[ImagePath]:
        """
        returns an iterator on ImagePath objects
        in this directory so that bool(predicate(image_path)) is True
        """
        directory = Path(directory)
        #2019 nov 22 - somehow this is broken
        #for filename in directory.glob(f"*{SUFFIX}"):
        for filename in glob.glob(f"{directory}/*{SUFFIX}"):
            image_path = ImagePath(self, filename)
            if predicate(image_path):
                yield image_path

    def locate_all_images(self, radical, look_in_global) -> List[ImagePath]:
        match = lambda image_path: (image_path.radical == radical
                                    or str(image_path) == radical)
        candidates = list(self._iterate_images(".", match))
        if look_in_global:
            candidates += list(self._iterate_images(self.public, match))
        candidates = [candidate for candidate in candidates if candidate.readable]
        candidates.sort(key=lambda info: info.mtime, reverse=True)
        return candidates


    def locate_image(self, radical, look_in_global) -> ImagePath:
        all_matches = self.locate_all_images(radical, look_in_global)
        return all_matches[0] if all_matches else None


    def _search_clusters(self, *, show_dot, show_public,
                         cluster_predicate, image_predicate,
                         sort_clusters=None, reverse=False):
        """
        search for images in either '.' and public dirs,
        that satisfy the image predicate
        then groups them by clusters
        filters out the ones that don't match
        sort the clusters
        displays them
        """
        candidates = []
        if show_dot:
            candidates += list(self._iterate_images(".", image_predicate))
        if show_public:
            candidates += list(self._iterate_images(
                self.public, image_predicate))
        clusters_hash = defaultdict(list)
        for image in candidates:
            if not image.readable:
                print(f"WARNING: ignoring unreadable image {image}")
                continue
            clusters_hash[image.inode].append(image)
        clusters = [ImageCluster(self, images)
                    for (inode, images) in clusters_hash.items()]
        clusters = [cluster for cluster in clusters
                    if cluster_predicate(cluster)]
        # sort_clusters is applied to ImageCluster objects
        if sort_clusters:
            clusters.sort(key=sort_clusters, reverse=reverse)
        return clusters


    def images(self, focus, sort_by, reverse,           # pylint: disable=r0913
               labeled, public_only, narrow):
        show_dot = not public_only
        show_public = True
        long_format = not narrow
        # is a given cluster filtered by focus
        def in_focus(cluster):
            # an empty focus means to show all
            if not focus:
                return True
            # in this context a cluster is just a list of infos
            for image_path in cluster:
                for filtr in focus:
                    # rough : no regexp, just find the filter or not
                    if str(image_path).find(filtr) >= 0:
                        return True
            return False
        def select_labeled(cluster):
            if not labeled:
                return True
            # else we select only the ones that have an alias
            return cluster.aliases >= 1
        def cluster_predicate(cluster):
            return select_labeled(cluster) and in_focus(cluster)
        def cluster_key(cluster):
            if sort_by == 'size':
                return cluster.regular.size
            elif sort_by == 'date':
                return cluster.regular.mtime
            elif sort_by == 'name':
                return cluster.regular.radical
            else:
                return 1

        clusters = self._search_clusters(
            show_dot=show_dot, show_public=show_public,
            cluster_predicate=cluster_predicate,
            image_predicate=lambda x: True,
            sort_clusters=cluster_key, reverse=reverse)
        # pylint: disable=w0212
        radical_width = max((cluster._max_radical() for cluster in clusters),
                            default=0)
        for cluster in clusters:
            # pylint: disable=w0212
            cluster._display(long_format, radical_width)


    def resolve(self, focus, verbose):
        """
        default output is just **one** readable path
        in verbose mode, the whole cluster(s) are displayed
        """
        def cluster_matches(cluster):
            return any(image.radical == focus
                       or str(image) == focus for image in cluster)

        true_function = lambda x: True
        clusters = self._search_clusters(
            show_dot=True, show_public=True,
            cluster_predicate=cluster_matches,
            image_predicate=true_function,
            sort_clusters=lambda cluster: cluster.regular.mtime, reverse=False)
        if not clusters:
            return 1
        if not verbose:
            print(clusters[-1].regular.path)
            return 0
        # verbose mode: list all clusters like in rimages
        # pylint: disable=w0212
        radical_width = max((cluster._max_radical() for cluster in clusters),
                            default=0)
        for cluster in clusters:
            # pylint: disable=w0212
            cluster._display(True, radical_width)
        return 0


    def share(self, image, alias,           # pylint:disable=r0912,r0913,r0914
              dry_run, force, clean) -> OsRetcod:
        """
        A utility to install one or several images in the shared repo

        Parameters:
          image: a local image name (not searched in global repo)
            to promote in the public repo
          alias: a name for the installed image

        image gets renamed into the radical from the origin name,
        i.e. with the extra saving__*blabla removed

        (.) identity
        root is allowed to override destination, not regular users
        thus this needs to be run under sudo, only privileged slices
        are so allowed

        # return 0 if all goes well, 1 otherwise
        """

        is_root = root_privileges()
        # compute dry_run if not set on the command line
        if dry_run is None:
            dry_run = not is_root
            if dry_run:
                print("WARNING: without sudo you can only dry-run")

        if not dry_run and not is_root:
            print("You need to run rhubarbe-share under sudo")
            return 1

        # if that's a filename, let's use it
        image_path = ImagePath(self, image)
        if not image_path.readable:
            image_path = self.locate_image(image, look_in_global=is_root)
        if not image_path:
            print(f"Could not find image {image} - ignored")
            return 1
        radical = image_path.radical
        matches = self.locate_all_images(radical, look_in_global=is_root)

        # first pass
        # compute a list of tuples (origin, destination)
        # all these objects are plain pathlib.Path instances
        moves = []     # list of tuples oldname, newname
        symlinks = []  # list of tuples plainfile, symlink
        removes = []
        chmods = []

        origin = image_path.path
        destination = self.public / (radical + SUFFIX)
        if destination.exists() and not force:
            print(f"WARNING: Destination {destination} already exists - ignored")
        else:
            moves.append((origin, destination))  # append a tuple
            chmods.append(destination)

        if alias:
            symlink = self.public / (alias + SUFFIX)
            symlinks.append((destination, symlink))
            chmods.append(symlink)
        else:
            print("WARNING: you are sharing an image without an alias")

        if clean:
            # item # 0 is the one selected for being moved
            for match in matches:
                if match == origin:
                    continue
                if match.is_official:
                    continue
                removes.append(match.path)

        matches.reverse()
        for index, match in enumerate(matches):
            prefix = '* ' if (index == len(matches) - 1) else '  '
            print(match._to_display(
                show_path=True, radical_width=len(radical), prefix=prefix))

        def show_dry_run(*args):
            print("DRY-RUN: would do: ", end="")
            print(*args)

        for remove in removes:
            if dry_run:
                show_dry_run(f"rm {remove}")
            else:
                print(f"Removing {remove}")
                remove.unlink()

        for origin, destination in moves:
            if dry_run:
                show_dry_run(f"mv {origin} {destination}")
            else:
                print(f"Moving {origin} -> {destination}")
                # the library folder may be shared over sshfs
                # and in that case using rename does not work 
                # because it is across filesystems
                # origin.rename(destination)
                command = f"mv {str(origin)} {str(destination)}"
                os.system(command)

        for plainfile, symlink in symlinks:
            if dry_run:
                show_dry_run(f"ln -sf {plainfile} {symlink}")
            else:
                print(f"Creating symlink {symlink} -> {plainfile}")
                if symlink.exists():
                    symlink.unlink()
                symlink.symlink_to(plainfile)

        for chmod in chmods:
            if dry_run:
                show_dry_run(f"chmod a+r {chmod}")
                continue
            omod = chmod.stat().st_mode
            nmod = omod | 0o444
            print(f"new mode for {chmod} is {oct(nmod)}")
            chmod.chmod(nmod)


        return 0



#!/usr/bin/env python3

import time

import asyncio
import telnetlib3
import logging

# if an individual command outputs more than that
# then we are in trouble
MAX_BUF = 16 * 1024

async def _shell(reader, writer, commands):

    recv = await reader.read(MAX_BUF)
    print(f'--init--(({recv}))', end='', flush=True)

    commands.append('echo _TELNET_STATUS=$?')
    commands.append('exit')

    for command in commands:
        print(f'[[{command}]]')
        writer.write(command + '\n')

    while True:
        if reader.at_eof():
            break
        recv = await reader.read(MAX_BUF)
        print(f"(({recv}))", end="", flush=True)

async def shell(reader, writer):
    commands = [
        'hostname',
        'cat /etc/fedora-release',
        'cat /etc/lsb-release',
        'frisbee -i 192.168.3.4 -m 234.5.6.1 -p 10001 /dev/sda',
    ]

    await _shell(reader, writer, commands)

async def complete(hostname, timeout, connect_minwait, connect_maxwait):
    print(f"timeout={timeout} min={connect_minwait}, max={connect_maxwait}")
    coro = telnetlib3.open_connection(
        hostname, 23, shell=None,
        connect_minwait=connect_minwait,
        connect_maxwait=connect_maxwait)
    start = time.time()
    try:
        result = reader, writer = await asyncio.wait_for(coro, timeout=timeout)
        middle = time.time()
        print(f"connection duration={middle-start}")
        await shell(reader, writer)
    except (asyncio.TimeoutError, OSError) as exc:
        # logging.error(f"OOPS {type(exc)} : {exc}")
        logging.error(f"OOPS: {type(exc)} {exc} ")
        result = None, None
    end = time.time()
    print(f"total duration={end-start}")
    return result


print(telnetlib3.__get_version())

from argparse import ArgumentParser

parser = ArgumentParser()
parser.add_argument("-t", "--timeout", dest='timeout', default=5.0, type=float)
parser.add_argument("-a", "--max", dest='connect_maxwait', default=3.0, type=float)
parser.add_argument("-i", "--min", dest='connect_minwait', default=1.0, type=float)
parser.add_argument("hostname")

args = parser.parse_args()

hostname = args.hostname

loop = asyncio.new_event_loop()
ok = loop.run_until_complete(
    complete(args.hostname, args.timeout,
             args.connect_minwait, args.connect_maxwait))
print(ok)



# pylint: disable=logging-fstring-interpolation
# pylint: disable=broad-exception-caught
# pylint: disable=missing-docstring

# pylint: disable=fixme

import json
import asyncio
import ssl

import websockets.uri
import websockets.protocol

from r2lab import SidecarAsyncClient

from rhubarbe.logger import monitor_logger as logger

class ReconnectableSidecar:

    def __init__(self, url, category, keep_period=1):
        # keep_period is the period at which connection is verified for open-ness
        self.url = url
        self.category = category
        self.keep_period = keep_period
        # caller MUST run keep_connected()
        self.connection = None
        self.counter = 0
        self.backlog = []
        logger.info(f"reconnectable sidecar to {url} ")

    def __repr__(self):
        size = f"no backlog" if not self.backlog else f"backlog={len(self.backlog)}"
        conn = f"no connection" if not self.connection else f"connection={self.connection.state}"
        return f"ReconnectableSidecar({self.url}) - {size} - {conn}"


    async def flush_backlog(self):
        # do not mess with the iteration subject
        for infos in self.backlog[:]:
            if await self.emit_infos(infos):
                self.backlog.remove(infos)


    # info singular: create a list with that one info
    async def emit_info(self, info):
        return await self.emit_infos([info])


    async def emit_infos(self, infos):
        # logger.debug(f"{self}: emit_infos is sending {infos}")
        if not self.connection:
            logger.warning(f"[no conn.] backlog ->  {infos}"
                           f" (with {len(self.backlog)} others)")
            self.backlog.append(infos)
            return False

        payload = dict(category=self.category, action='info', message=infos)
        try:
            await self.connection.send(json.dumps(payload))
            self.counter += 1
            return True
        except ConnectionRefusedError:
            logger.warning(f"[conn. refused] message {infos} goes into backlog"
                           f" (with {len(self.backlog)} others)")
            self.backlog.append(infos)
            return False
        except websockets.exceptions.ConnectionClosedError as exc:
            logger.warning(f"[conn. closed] message {infos} goes into backlog"
                           f" (with {len(self.backlog)} others)")
            self.backlog.append(infos)
            self.connection = None
            return False
        except Exception as exc:
            # xxx to review
            logger.exception(f"connection.send failed: {type(exc)}: {exc}")
            self.connection = None
            return False
        finally:
            pass
            # logger.debug(f"<<< emit_infos is leaving connection.send() with {payload}")


    async def keep_connected(self):
        """
        A continuous loop that keeps the connection open
        """
        while True:
            if self.connection and self.connection.state == websockets.protocol.State.OPEN:
                logger.debug(f"in keep_connected, connection is open")
                if self.backlog:
                    logger.info(f"flushing backlog of {len(self.backlog)} messages")
                    await self.flush_backlog()
                    logger.info(f"after flush, backlog now has {len(self.backlog)} messages")
            else:
                # xxx should we close() our client ?
                self.connection = None
                # see if we need ssl
                secure = websockets.uri.parse_uri(self.url).secure
                kwds = {}
                if secure:
                    kwds.update(dict(ssl=ssl.SSLContext()))
                try:
                    logger.info(f"(re)-connecting to {self.url} ...")
                    self.connection = await SidecarAsyncClient(self.url, **kwds)
                    logger.debug("connected !")
                except ConnectionRefusedError:
                    logger.warning(f"Could not connect to {self.url} at this time")
                except Exception as exc:
                    logger.exception(
                        f"Could not connect to {self.url} at this time - uncaught {exc}")
            await asyncio.sleep(self.keep_period)


    async def watch_back_channel(self, category, callback):
        while True:
            if not self.connection:
                logger.debug(f"backing off for {self.keep_period}s")
                await asyncio.sleep(self.keep_period)
                continue
            try:
                incoming = await self.connection.recv()
                umbrella = json.loads(incoming)
                logger.info(f"tmp - got incoming {umbrella['category']} x {umbrella['action']}")
                if (umbrella['category'] == category and
                        umbrella['action'] == 'request'):
                    callback(umbrella)
            except Exception as exc:
                ### to review
                logger.exception(f"recv failed .. fix me {exc}")
                self.connection = None



"""
The SshProxy class is the very early seed of what became SshProxy in apssh

This code still has it, mostly out of laziness
"""

# c0111 no docstrings yet
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# pylint: disable=c0111, w0703, w1202

import time
import random
import asyncio
import asyncssh

DEBUG = False
# DEBUG = True

from .logger import monitor_logger as logger


class MySSHClientSession(asyncssh.SSHClientSession):
    """
    this class is a specialization of asyncssh's session object,
    - see clientsession_closure below -
    this is how we can access self.node which is a reference to
    the corresponding Node object
    """
    def __init__(self, *args, **kwds):
        self.data = ""
        self.node = None
        self.command = None
        super().__init__(*args, **kwds)

    def data_received(self, data, datatype):
        # not adding a \n since it's already in there
        if DEBUG:
            print('SSS DR: {}:{}-> {} [[of type {}]]'.
                  format(self.node, self.command, data, datatype), end='')
        self.data += data

    def connection_made(self, conn):                    # pylint: disable=w0221
        if DEBUG:
            print(f'SSS CM: {self.node} {conn}')

    def connection_lost(self, exc):
        if exc:
            if DEBUG:
                print(f'SSS CL: {self.node} - exc={exc}')

    def eof_received(self):
        if DEBUG:
            print(f'SSS EOF: {self.node}')


class MySSHClient(asyncssh.SSHClient):
    def connection_made(self, conn):
        if DEBUG:
            print(f"SSC Connection made to "
                  f" {conn.get_extra_info('peername')[0]}.")

    def auth_completed(self):
        if DEBUG:
            print('SSC Authentication successful.')


class SshProxy:
    """
    talk to a Node's control interface using ssh
    """
    def __init__(self, node, username='root', verbose=False):
        self.node = node
        self.username = username
        self.verbose = verbose
        #
        self.hostname = self.node.control_hostname()
        self.status = None
        self.conn, self.client = None, None

    def __repr__(self):
        return f"SshProxy {self.node}"

    # make this an asynchroneous context manager
    # async with SshProxy(...) as ssh:
    #
    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_value, traceback):
        # xxx this might be a little harsh, in the case
        # where an exception did occur
        await self.close()

    async def connect(self, timeout=None):
        begin = time.time()
        try:
            self.conn, self.client = await asyncio.wait_for(
                asyncssh.create_connection(
                    MySSHClient, self.hostname, username=self.username,
                    known_hosts=None
                ),
                timeout=timeout)
            retcod = True
        except (OSError, asyncssh.Error, asyncio.TimeoutError, asyncio.exceptions.CancelledError) as exc:
            logger.info(f"SSH FAIL on {self.hostname} {type(exc)=} {exc=}")
            self.conn, self.client = None, None
            retcod = False
        except ValueError as exc:
            # seen raised by asyncssh for some reason,
            # anyway bottom line is we can't connect
            #
            logger.info(f"SSH FAIL on {self.hostname} with ValueError, {exc=}")
            self.conn, self.client = None, None
            retcod = False
        finally:
            end = time.time()
            logger.info(f"SSH connect {self.hostname} took {end-begin:.3f}s {retcod=}")
            return retcod

    async def run(self, command):
        """
        Run a command
        """
        class ClientsessionClosure(MySSHClientSession):
            def __init__(ssh_client_session,            # pylint: disable=e0213
                         *args, **kwds):
                ssh_client_session.node = self.node
                ssh_client_session.command = command
                super().__init__(*args, **kwds)

        # print(5*'-', "running on ", self.hostname, ':', command)
        begin = time.time()
        try:
            chan, session = await self.conn.create_session(
                ClientsessionClosure, command)
            await chan.wait_closed()
            output = session.data
        except Exception as exc:
            logger.info(f"failed to SSH run {self.hostname} - {type(exc)=} {exc=}")
            output = None
        finally:
            end = time.time()
            logger.info(
                f"SSH run {self.hostname} took {end-begin:.3f}s")
            return output

    # >>> asyncio.iscoroutine(asyncssh.SSHClientConnection.close)
    # False
    async def close(self):
        if self.conn is not None:
            self.conn.close()
            await self.conn.wait_closed()
        self.conn = None

    async def wait_for(self, backoff, timeout=1.):
        """
        Wait until the ssh service is usable
        """
        self.status = False
        while True:
            if self.verbose:
                await self.node.feedback('ssh_status', "trying to connect")
            self.status = await self.connect(timeout)
            if self.status:
                if self.verbose:
                    await self.node.feedback('ssh_status', "connection OK")
                await self.close()
                return self.status
            # random.random() is between 0. and 1.
            # and so as we need something between 0.5 and 1.5
            random_backoff = (0.5 + random.random()) * backoff
            if self.verbose:
                await self.node.feedback(
                    'ssh_status',
                    f"cannot connect, backing off for {random_backoff:.3}s")
            await asyncio.sleep(random_backoff)


# mostly test-oriented
if __name__ == '__main__':

    def main():
        import sys
        from rhubarbe.node import Node

        async def probe(host, message_bus):
            node = Node(host, message_bus)
            proxy = SshProxy(node, verbose=True)
            conn = await proxy.connect()
            if not conn:
                return False
            out1 = await proxy.run(
                'cat /etc/lsb-release /etc/fedora-release 2> /dev/null')
            print(f"command1 returned {out1}")
            out2 = await proxy.run('hostname')
            print(f"command2 returned {out2}")
            await proxy.close()
            return True

        message_bus = asyncio.Queue()

        nodes = sys.argv[1:]
        tasks = [probe(node, message_bus) for node in nodes]

        retcods = asyncio.new_event_loop()\
            .run_until_complete(asyncio.gather(*tasks))

        for node, retcod in zip(nodes, retcods):
            print(f"{node}:{retcod}")
    main()



import asyncio
import aioping

import sys
target = sys.argv[1]

async def do_ping1(host):
    print(f"{host=}")
    try:
        delay = await aioping.ping(host, timeout=1) * 1000
        print("Ping response in %s ms" % delay)

    except TimeoutError:
        print("Timed out")


async def do_ping2(host):
    command = ["ping", "-c", "1", "-t", "1", host]
    ping_timeout = 0.5
    subprocess = await asyncio.create_subprocess_exec(
        *command,
        stdout=asyncio.subprocess.DEVNULL,
        stderr=asyncio.subprocess.DEVNULL)
    # failure occurs through timeout
    returncode = await asyncio.wait_for(
        subprocess.wait(), timeout=ping_timeout)
    print(f"{returncode=}")

print("with aioping")
asyncio.run(do_ping1(target))
print("with ping & subprocess")
asyncio.run(do_ping2(target))



"""
display class when not using curses, like in
rhubarbe load
"""

import time

# pip3 install progressbar33
import progressbar

from rhubarbe.logger import logger

# c0111 no docstrings yet
# w0201 attributes defined outside of __init__
# w1202 logger & format
# w0703 catch Exception
# r1705 else after return
# r0913 too many arguments in functions
# pylint: disable=c0111,w1202,w0201,r1705,r0913


# message_bus is just an asyncio.Queue

# a display instance comes with a hash
# 'ip' -> DisplayNode
# in its simplest form a DisplayNode just has
# a hostname (retrieved from ip through inventory)
# a rank in the nodes list (starting at 0)
# a percentage

class DisplayNode:                                      # pylint: disable=r0903
    def __init__(self, name, rank):
        self.name = name
        self.rank = rank
        self.percent = 0


class Display:                                          # pylint: disable=r0902
    def __init__(self, nodes, message_bus):
        self.message_bus = message_bus
        self.nodes = nodes
        #
        self._alive = True
        self._start_time = None
        # this will go from 0 to 100*len(self.nodes)
        self.total_percent = 0
        # correspondance ip -> display_node
        self._display_node_by_ip = {}
        self.goodbye_message = None
        # for the basic displaying : we use a ingle global progress bar
        self.pbar = None

    def get_display_node(self, ipaddr):
        # if we have it already
        if ipaddr in self._display_node_by_ip:
            return self._display_node_by_ip[ipaddr]

        # in case the incoming ip is a reboot ip
        from rhubarbe.inventorynodes import InventoryNodes
        the_inventory = InventoryNodes()
        control_ip = the_inventory.control_ip_from_any_ip(ipaddr)
        # locate this in the subject nodes list
        for rank, node in enumerate(self.nodes):
            if node.control_ip_address() == control_ip:
                self._display_node_by_ip[ipaddr] = \
                    DisplayNode(node.control_hostname(), rank)
                return self._display_node_by_ip[ipaddr]
        return None

    async def run(self):
        self.start_hook()
        self._start_time = time.time()

        while self._alive:
            message = await self.message_bus.get()
            if message == 'END-DISPLAY':
                self._alive = False
                break
            self.dispatch(message)
            # this is new in 3.4.4
            if 'task_done' in dir(self.message_bus):
                self.message_bus.task_done()

    async def stop(self):
        # soft stop
        await self.message_bus.put("END-DISPLAY")

#    def stop_nowait(self):
#        if self._alive:
#            self._alive = False
#            self.stop_hook()

    def dispatch(self, message):
        timestamp = time.strftime("%H:%M:%S")
        # in case the message is sent before the event loop has started
        duration = (f"+{int(time.time()-self._start_time):03}s"
                    if self._start_time is not None else 5*'-')
        if isinstance(message, dict) and 'ip' in message:
            ipaddr = message['ip']
            node = self.get_display_node(ipaddr)
            if node is None:
                logger.info(f"Unexpected message gave node=None in dispatch: {message}")
            elif 'tick' in message:
                self.dispatch_ip_tick_hook(ipaddr, node, message,
                                           timestamp, duration)
            elif 'percent' in message:
                # compute delta, update node.percent and self.total_percent
                node_previous_percent = node.percent
                node_current_percent = message['percent']
                delta = node_current_percent - node_previous_percent
                node.percent = node_current_percent
                self.total_percent += delta
                logger.info(f"{node.name} percent: {node_current_percent}/100 "
                            f"(was {node_previous_percent}), "
                            f"total {self.total_percent}/{100*len(self.nodes)}")
                self.dispatch_ip_percent_hook(ipaddr, node, message,
                                              timestamp, duration)
            else:
                self.dispatch_ip_hook(ipaddr, node, message,
                                      timestamp, duration)
        else:
            self.dispatch_hook(message, timestamp, duration)

    @staticmethod
    def message_to_text(message):
        if isinstance(message, str):
            return message
        elif not isinstance(message, dict):
            # should not happen
            return "UNEXPECTED" + str(message)
        elif 'info' in message:
            return message['info']
        elif 'authorization' in message:
            return "AUTH: " + message['authorization']
        elif 'loading_image' in message:
            return f"Loading image {message['loading_image']}"
        elif 'selected_nodes' in message:
            names = message['selected_nodes'].node_names()
            return ("Selection: " + " ".join(names)) \
                if names \
                else "Empty Node Selection"
        else:
            return str(message)

    subkeys = ['frisbee_retcod', 'reboot', 'ssh_status', 'frisbee_status']

    def message_to_text_ip(self, message, node, mention_node=True):
        text = None
        if 'percent' in message:
            text = f"{message['percent']:02}"
        elif 'frisbee_retcod' in message:
            text = "Uploading successful" \
                if message['frisbee_retcod'] == 0 \
                else "Uploading FAILED !"
        else:
            for key in self.subkeys:
                if key in message:
                    text = f"{key} = {message[key]}"
                    break
        if text is None:
            text = str(message)
        return text \
            if not mention_node \
            else f"{node.name} : {text}"

    def set_goodbye(self, message):
        self.goodbye_message = message

    ####################
    # specifics of the basic display
    def start_hook(self):
        pass

    def epilogue(self):
        if self.goodbye_message:
            print(self.goodbye_message)

    def repair(self):
        self.epilogue()

    def dispatch_hook(self, message, timestamp, duration):
        text = self.message_to_text(message)
        print(f"{timestamp} - {duration}: {text}")

    def dispatch_ip_hook(self, ipaddr, node,            # pylint: disable=w0613
                         message, timestamp, duration):
        text = self.message_to_text_ip(message, node, mention_node=False)

        print(f"{timestamp} - {duration}: {node.name} {text}")

    def dispatch_ip_percent_hook(self, ipaddr, node,    # pylint: disable=w0613
                                 message, timestamp,    # pylint: disable=w0613
                                 duration):             # pylint: disable=w0613
        # start progressbar
        if self.pbar is None:
            widgets = [
                progressbar.Bar(),
                progressbar.Percentage(), ' |',
                progressbar.FormatLabel('%(seconds).2fs'), '|',
                progressbar.ETA(),
            ]
            self.pbar = \
                progressbar.ProgressBar(widgets=widgets,
                                        maxval=len(self.nodes)*100)
            self.pbar.start()
        self.pbar.update(self.total_percent)
        if self.total_percent == len(self.nodes)*100:
            self.pbar.finish()

    def dispatch_ip_tick_hook(self, ipaddr, node,       # pylint: disable=w0613
                              message, timestamp,       # pylint: disable=w0613
                              duration):                # pylint: disable=w0613
        # start progressbar
        if self.pbar is None:
            widgets = [
                'Collecting image : ',
                # progressbar.BouncingBar(marker=progressbar.RotatingMarker()),
                progressbar.BouncingBar(marker='*'),
                progressbar.FormatLabel(' %(seconds).2fs'),
            ]
            self.pbar = \
                progressbar.ProgressBar(widgets=widgets,
                                        maxval=progressbar.UnknownLength)
            self.pbar.start()
            # progressbar is willing to work as expected here
            # with maxval=UnknownLength
            # but still insists to get a real value apparently
            self.value = 0
        self.value += 1
        self.pbar.update(self.value)
        # hack way to finish the progressbar
        # since we have no other way to figure it out
        if message['tick'] == 'END':
            self.pbar.finish()



"""
a utility to check if the testbed is free for a given time slot
"""

# could have been inserted into lease.py, but that one is already too big
# and convoluted

import time
from datetime import datetime as DateTime
import argparse

from .config import Config
from .plcapiproxy import PlcApiProxy

class Book:

    def __init__(self, email=None, password=None, verbose=False):
        self._verbose = verbose
        plcapi_url = Config().value('plcapi', 'url')
        self.plcapi_proxy = PlcApiProxy(plcapi_url, email=email, password=password)
        self.leases_hostname = Config().value('plcapi', 'leases_hostname')

    def verbose(self, *args, **kwargs):
        if self._verbose:
            print(*args, **kwargs)


    def leases(self, start, end) -> list["PlcLease"]:
        """
        return the leases for the given time slot
        """
        leases = self.plcapi_proxy.GetLeases({'clip': (start, end)})
        # sort by start time
        leases.sort(key=lambda lease: lease['t_from'])
        return leases

    def canonical_date(self, date):
        """
        return a date in the format expected by the plcapi
        tries to be flexible and to understand various formats
        """
        # if the date is given as just hh:mm, assume it is today
        if len(date) == 5:
            date = f"{DateTime.now().strftime('%Y-%m-%d')}T{date}"
        formats = [
            '%Y-%m-%dT%H:%M',
            '%Y-%m-%d %H:%M',
        ]
        local = None
        for format in formats:
            try:
                local = DateTime.strptime(date, format)
                break
            except ValueError:
                pass
        if local is None:
            raise ValueError(f"cannot parse date {date}")
        epoch = int((local - DateTime(1970, 1, 1)).total_seconds())

        # translate to UTC
        is_dst = time.daylight and time.localtime().tm_isdst > 0
        utc_offset = (time.altzone if is_dst else time.timezone)
        return epoch + utc_offset

    def date_to_string(self, date):
        """
        return a date in the format expected by the plcapi
        """
        return DateTime.fromtimestamp(date).strftime('%Y-%m-%dT%H:%M')

    def lease_repr(self, lease):
        """
        return a string representation of the lease
        """
        return (
            f"{self.date_to_string(lease['t_from'])} -> "
            f"{self.date_to_string(lease['t_until'])} "
            f"{lease['name']} "
        )

    def query(self, start, end) -> bool:
        """
        return True if the time slot is free
        """
        leases = self.leases(start, end)
        self.verbose(
            f"{len(leases)} leases from {self.date_to_string(start)} "
            f"-> {self.date_to_string(end)}")
        for lease in leases:
            self.verbose(f"{14*' '}{self.lease_repr(lease)}")
        return not leases

    def book(self, slice, start, end) -> bool:
        """
        book the time slot - return True if successful
        """
        self.verbose(
            f"booking {slice} from {self.date_to_string(start)} "
            f"until {self.date_to_string(end)}")
        hostname = self.leases_hostname
        try:
            retcod = self.plcapi_proxy.AddLeases(
                    [hostname], slice, start, end
            )
            if retcod['errors']:
                for error in retcod['errors']:
                    print(f"error: {error}")
                return False
            else:
                lease_id = retcod['new_ids'][0]
                self.verbose(f"new lease id: {lease_id}")
                return True
        except Exception as exc:
            print(f"exception in AddLeases : {type(exc)}: {exc}")
            return False

    def delete(self, start, end) -> bool:
        """
        spots the - expected single - lease during the time slot
           and deletes it
        implements the same logic as the website, i.e.:
        - a lease in the past is not deleted
        - a currently running lease is updated/shortened so as to release
          the testbed but still keep track of that lease
        - a future lease - or a shortly currently lease - is deleted

        return True if everything went well
        """

        # find the lease
        leases = self.leases(start, end)
        if len(leases) != 1:
            print(f"error: from {self.date_to_string(start)} "
                f"until {self.date_to_string(end)}")
            print(f"expected a single lease, found {len(leases)} leases")
            return False

        lease = leases[0]
        lease_id = lease['lease_id']
        now = time.time()

        # if the lease is in the past, do nothing
        if lease['t_until'] < now:
            self.verbose(f"won't delete lease in the past")
            return False

        # how long has it been running ?
        granularity = self.plcapi_proxy.GetLeaseGranularity()
        started = now - lease['t_from']
        # it's been running for more than one granularity, update it
        if started > granularity:
            # we keep that many grains
            nb_grains = int(started // granularity)
            new_end = lease['t_from'] + nb_grains * granularity
            self.verbose(
                f"updating lease {lease_id} until {self.date_to_string(new_end)}")
            try:
                retcod = self.plcapi_proxy.UpdateLeases(
                    [lease_id], {'t_until': new_end})
                if retcod['errors']:
                    for error in retcod['errors']:
                        print(f"error: {error}")
                    return False
                else:
                    return True
            except Exception as exc:
                print(f"exception in UpdateLease : {type(exc)}: {exc}")
                return False

        # still here ? the lease has not yet started,
        # or it has been running for less than one granularity
        # delete it
        self.verbose(f"deleting future lease {lease_id}")
        try:
            retcod = self.plcapi_proxy.DeleteLeases([lease_id])
            if retcod == 1:
                self.verbose(f"successful")
                return True
            print(f"unexpected error {retcod=} while deleting {lease_id=}")
            return False
        except Exception as exc:
            print(f"exception in DeleteLease : {type(exc)}: {exc}")
            return False

    @staticmethod
    def main(*argv) -> bool:
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-q", "--query", action="store_true",
            help="does not book, just checks if the slot is free")
        parser.add_argument(
            "-d", "--delete", action="store_true",
            help="delete the lease for the given time slot")
        parser.add_argument(
            "-s", "--slice",
            help="slice name - mandatory unless using -q")
        parser.add_argument(
            "-e", "--email", help="email")
        parser.add_argument(
            "-p", "--password", help="password")
        parser.add_argument(
            "-v", "--verbose", action="store_true",
            help="be verbose")
        parser.add_argument(
            "start",
            help="start time format: 2020-01-01T00:00 or 2020-01-01 00:00 or 14:00")
        parser.add_argument(
            "end",
            help="end time, same format as start time")
        args = parser.parse_args(*argv)
        book = Book(
            email=args.email, password=args.password, verbose=args.verbose)
        start = book.canonical_date(args.start)
        end = book.canonical_date(args.end)
        if args.query:
            return book.query(start, end)
        elif args.delete:
            return book.delete(start, end)
        else:
            if args.slice is None:
                parser.error("slice is mandatory unless using -q")
                exit(1)
            return book.book(args.slice, start, end)



#!/usr/bin/env python3

# issuing GetNodes() on our plcapi instance

url = "https://r2labapi.inria.fr:443/PLCAPI/"
auth = { 'AuthMethod' : 'password',
         'Username'   : 'root@r2lab.inria.fr',
         'AuthString' : 'onecalvin' }

import ssl

context = ssl.SSLContext(ssl.PROTOCOL_TLSv1)
context.check_hostname = False


########## synchroneous version

def sync_nodes():

    import xmlrpc.client

    proxy = xmlrpc.client.ServerProxy(url, allow_none=True, context=context)

    nodes = proxy.GetNodes(auth, {}, ['hostname'] )

    print("synchro nodes[0] -> {}".format(nodes[0]))


########## async

import asyncio


async def co_async_nodes():

    import aioxmlrpc.client

    try:
        proxy = aioxmlrpc.client.ServerProxy(url, allow_none=True, context=context)
    except:
        import traceback
        traceback.print_exc()
        exit(1)
    nodes = await proxy.GetNodes(auth, {}, ['hostname'] )
    print("asynchro nodes[0] -> {}".format(nodes[0]))


def async_nodes():
    loop = asyncio.new_event_loop()
    loop.run_until_complete(co_async_nodes())


def main():
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("-s", "--sync", default=False, action='store_true')
    args = parser.parse_args()
    synchro_mode = args.sync

    if synchro_mode:
        sync_nodes()
    else:
        async_nodes()

main()
